import argparse
import logging
from pathlib import Path
import os
import re

import requests

logger = logging.getLogger(__name__)

SEATTLE_DMR_REPEATERS = "http://seattledmr.org/ConfigBuilder/Digital-Repeaters-Seattle-addon.csv"
SEATTLE_DMR_TALKGROUPS = "http://seattledmr.org/ConfigBuilder/Talkgroups-Seattle-addon.csv"
REPEATER_FILENAME = "Digital-Repeaters__SeattleDMR.csv"
TALKGROUPS_FILENAME = "Talkgroups__SeattleDMR.csv"


def cache_repeaters(output_dir):
    repeaters = requests.get(SEATTLE_DMR_REPEATERS)
    repeaters.raise_for_status()
    talkgroups = requests.get(SEATTLE_DMR_TALKGROUPS)
    talkgroups.raise_for_status()
    outpath = Path(output_dir)
    rp_out = outpath / REPEATER_FILENAME
    with rp_out.open("w", newline='') as f:
        # XXX: Hacks: need to fix upstream
        for line in repeaters.text.splitlines(True):
            line = line.replace("BayNet", "Baynet").replace("PNWR", "PNW Rgnl 2")
            line = line.replace("Wash 1", "Washington 1").replace("Wash 2", "Washington 2")
            f.write(line)
    logger.info("Cache SeattleDMR k7abd zones to '%s'", rp_out)

    tg_out = outpath / TALKGROUPS_FILENAME
    with tg_out.open("w", newline='') as f:
        # XXX: Hacks: need to fix upstream
        for line in talkgroups.text.splitlines(True):
            line = re.sub(r"Link([0-9]+)", r"Link \1", line)
            f.write(line)
        f.write("TAC 8-2,8958\n")
    logger.info("Cache SeattleDMR k7abd talkgroups to '%s'", tg_out)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("output_dir")
    args = parser.parse_args()
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    cache_repeaters(args.output_dir)



"""
dzcb main program
"""

import argparse
import json
from importlib_resources import contents, files
import logging
from pathlib import Path
import os
import shutil

from dzcb import __version__
import dzcb.anytone
import dzcb.data
import dzcb.farnsworth
import dzcb.gb3gf
import dzcb.k7abd
import dzcb.log
import dzcb.model
import dzcb.pnwdigital
import dzcb.repeaterbook
import dzcb.seattledmr


logger = logging.getLogger("dzcb")


def append_dir_and_create(path, component=None):
    new_path = path
    if component:
        new_path = new_path / component
    new_path.mkdir(parents=True, exist_ok=True)
    return new_path


def repeaterbook_proximity_csv(csv_file, cache_dir):
    proximity_csv = csv_file.read().splitlines()
    rbcache = append_dir_and_create(cache_dir, "repeaterbook")
    dzcb.repeaterbook.cache_zones_with_proximity(proximity_csv, rbcache)
    dzcb.repeaterbook.zones_to_k7abd(proximity_csv, rbcache, cache_dir)


def cache_user_or_default_json(object_name, user_path, default_path, cache_dir):
    """
    Read JSON from a user-specified or default path for object_name.

    Side-effects:
      * Logging at info level, mentioning object_name and the path used
      * Copying either the user_path or default_path to cache_dir

    Return:
        Python objects
    """
    if user_path is None:
        logger.info("Cache default %s: '%s'", object_name, default_path)
        path = default_path
    else:
        logger.info("Cache user-specified %s: '%s'", object_name, user_path)
        path = Path(user_path)
    dest = cache_dir / path.name
    shutil.copy(path, dest)
    return json.loads(dest.read_text())


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="dzcb: DMR Zone Channel Builder")
    parser.add_argument(
        "--pnwdigital",
        action="store_true",
        help="Fetch the latest pnwdigital K7ABD input files",
    )
    parser.add_argument(
        "--seattledmr",
        action="store_true",
        help="Fetch the latest seattledmr K7ABD input files",
    )
    parser.add_argument(
        "--repeaterbook-proximity-csv",
        type=argparse.FileType("r"),
        default=None,
        help="Fetch repeaters within X distance of POIs defined in a CSV file",
    )
    parser.add_argument(
        "--repeaterbook-state",
        nargs="*",
        help="Download repeaters from the given state(s). Default: '{}'".format(
            "' '".join(dzcb.repeaterbook.REPEATERBOOK_DEFAULT_STATES),
        ),
    )
    parser.add_argument(
        "--default-k7abd",
        action="store_true",
        help="Include bundled K7ABD input files (simplex + unlicensed)",
    )
    parser.add_argument(
        "--k7abd",
        nargs="*",
        help="Specify one or more local directories containing K7ABD CSV files",
    )
    parser.add_argument(
        "--farnsworth-template-json",
        nargs="*",
        help="JSON file to take Farnsworth settings from. If no json file, defaults will "
        "be used for each supported radio type.",
    )
    parser.add_argument(
        "--scanlists-json",
        default=None,
        help="JSON dict mapping scanlist name to list of channel names.",
    )
    parser.add_argument(
        "--order-json",
        default=None,
        help="JSON dict specifying zone and talkgroup orderings.",
    )
    parser.add_argument("outdir", help="Write code plug files to this directory")
    args = parser.parse_args()

    outdir = append_dir_and_create(Path(args.outdir))
    dzcb.log.init_logging(log_path=outdir)
    logger.info("dzcb %s outdir='%s'", __version__, outdir)

    cache_dir = append_dir_and_create(outdir, "cache")
    logger.debug("Using cache_dir: '%s'", cache_dir)

    # fetch data from the internet
    if args.repeaterbook_proximity_csv:
        dzcb.repeaterbook.zones_to_k7abd(
            input_csv=args.repeaterbook_proximity_csv,
            output_dir=cache_dir,
            states=args.repeaterbook_state or dzcb.repeaterbook.REPEATERBOOK_DEFAULT_STATES,
        )

    if args.pnwdigital:
        dzcb.pnwdigital.cache_repeaters(cache_dir)

    if args.seattledmr:
        dzcb.seattledmr.cache_repeaters(cache_dir)

    if args.default_k7abd:
        default_k7abd_path = files(dzcb.data) / "k7abd"
        logger.info("Cache default k7abd zones from: '%s'", default_k7abd_path)
        shutil.copytree(default_k7abd_path, cache_dir, dirs_exist_ok=True)

    if args.k7abd:
        # copy any additional CSV directories into the cache_dir
        for abd_dir in args.k7abd:
            logger.info("Cache k7abd zones from: '%s'", abd_dir)
            shutil.copytree(abd_dir, cache_dir, dirs_exist_ok=True)

    # load additional data files or defaults
    scanlists = cache_user_or_default_json(
        object_name="scanlists",
        user_path=args.scanlists_json,
        default_path=files(dzcb.data).joinpath("scanlists.json"),
        cache_dir=cache_dir,
    )
    order = cache_user_or_default_json(
        object_name="zone order",
        user_path=args.order_json,
        default_path=files(dzcb.data).joinpath("order.json"),
        cache_dir=cache_dir,
    )
    zone_order = order.get("zone", {}).get("default", [])
    zone_order_expanded = order.get("zone", {}).get("expanded", [])
    exclude_zones_expanded = order.get("zone", {}).get("exclude_expanded", [])
    static_talkgroup_order = order.get("static_talkgroup", [])

    # create codeplug from a directory of k7abd CSVs
    cp = (
        dzcb.k7abd.Codeplug_from_k7abd(cache_dir)
        .order_grouplists(static_talkgroup_order=static_talkgroup_order)
        .replace_scanlists(scanlists)
    )
    logger.info("Generated %s", cp)

    # GB3GF CSV - Radioddity GD77/OpenGD77, TYT MD-380, MD-9600, Baofeng DM1801, RD-5R
    # XXX: Only support OpenGD77 at the moment
    gb3gf_outdir = append_dir_and_create(outdir, "gb3gf")
    gd77_outdir = append_dir_and_create(gb3gf_outdir, "opengd77")
    dzcb.gb3gf.Codeplug_to_gb3gf_opengd77_csv(
        cp.order_zones(zone_order=zone_order),
        output_dir=gd77_outdir,
    )

    # The following models use expand_static_talkgroups to create
    # one channel per talkgroup / one zone per repeater
    fw_cp = (
        cp.expand_static_talkgroups(static_talkgroup_order=static_talkgroup_order)
        .order_zones(
            zone_order=zone_order_expanded,
            exclude_zones=exclude_zones_expanded,
        )
        .replace_scanlists(scanlists)
    )
    logger.info("Expand static talkgroups %s", fw_cp)

    # Anytone 578/868/878 stock CPS CSV import files
    anytone_outdir = append_dir_and_create(outdir, "anytone")
    dzcb.anytone.Codeplug_to_anytone_csv(
        cp=fw_cp,
        output_dir=anytone_outdir,
    )

    # Farnsworth JSON - TYT et. al w/ Zone Import!
    farnsworth_templates = []
    if args.farnsworth_template_json is None:
        # Iterate through all farnsworth templates, generating codeplug for each
        for f in (files(dzcb.data) / "farnsworth").iterdir():
            if not str(f).endswith(".json"):
                continue
            farnsworth_templates.append((f, f.name, f.open("r")))
    else:
        farnsworth_templates = [
            (ftj, os.path.basename(ftj), open(ftj, "r"))
            for ftj in args.farnsworth_template_json
        ]

    fw_outdir = append_dir_and_create(outdir, "editcp")
    for ftj, fname, fh in farnsworth_templates:
        outfile = fw_outdir / fname
        outfile.write_text(
            dzcb.farnsworth.Codeplug_to_json(
                cp=fw_cp,
                based_on=fh,
            )
        )
        logger.info("Wrote '%s' based JSON to '%s'", ftj, outfile)



"""
dzcb.model - data model for codeplug objects
"""
import enum
import logging

import attr

# cached data for testing
import dzcb.data
import dzcb.exceptions
import dzcb.munge

# XXX: i hate this
NAME_MAX = 16

logger = logging.getLogger(__name__)


class ConvertibleEnum(enum.Enum):
    @classmethod
    def from_any(cls, v):
        """Passable as an attr converter."""
        if isinstance(v, cls):
            return v
        return cls(v)

    def __str__(self):
        return str(self.value)


class Timeslot(ConvertibleEnum):
    """
    DMR Tier II uses 2x30ms timeslots capable of carrying independent traffic.
    """

    ONE = 1
    TWO = 2

    @classmethod
    def from_any(cls, v):
        """Passable as an attr converter."""
        if isinstance(v, cls):
            return v
        return cls(int(v))


class ContactType(ConvertibleEnum):
    GROUP = "Group"
    PRIVATE = "Private"


@attr.s(eq=True, frozen=True)
class Contact:
    """
    A Digital Contact: group or private
    """

    _all_contacts_by_id = {}

    name = attr.ib(eq=True, converter=dzcb.munge.contact_name)
    dmrid = attr.ib(eq=True, order=True, converter=int)
    kind = attr.ib(
        eq=True,
        default=ContactType.GROUP,
        validator=attr.validators.instance_of(ContactType),
        converter=ContactType.from_any,
    )

    def __attrs_post_init__(self):
        all_contacts_by_id = type(self)._all_contacts_by_id
        existing_contact = all_contacts_by_id.get(self.dmrid, None)
        if existing_contact:
            raise dzcb.exceptions.DuplicateDmrID(
                msg="{} ({}): DMR ID already exists as {}".format(
                    self.name, self.dmrid, existing_contact.name
                ),
                existing_contact=existing_contact,
            )
        all_contacts_by_id[self.dmrid] = self


@attr.s(eq=True, frozen=True)
class Talkgroup(Contact):
    timeslot = attr.ib(
        eq=True,
        order=True,
        default=Timeslot.ONE,
        validator=attr.validators.instance_of(Timeslot),
        converter=Timeslot.from_any,
    )

    def __attrs_post_init__(self):
        try:
            super().__attrs_post_init__()
        except dzcb.exceptions.DuplicateDmrID as dup:
            if (
                isinstance(dup.existing_contact, type(self))
                and self.timeslot == dup.existing_contact.timeslot
            ):
                raise

    @property
    def name_with_timeslot(self):
        ts = str(self.timeslot)
        if self.name.endswith(ts) and not self.name.startswith("TAC"):
            name = self.name
        else:
            name = "{} {}".format(self.name, ts)
        return name

    @classmethod
    def from_contact(cls, contact, timeslot):
        fields = attr.asdict(contact)
        fields["timeslot"] = timeslot
        try:
            return cls(**fields)
        except dzcb.exceptions.DuplicateDmrID as dup:
            return dup.existing_contact  # return the talkgroup we already have


class Power(ConvertibleEnum):
    LOW = "Low"
    MED = "Medium"
    HIGH = "High"


@attr.s(frozen=True)
class Channel:
    """Common channel attributes"""

    # keep track of short names to avoid duplicate names
    _all_channels = {}
    _short_names = {}

    name = attr.ib(validator=attr.validators.instance_of(str))
    frequency = attr.ib(validator=attr.validators.instance_of(float), converter=float)
    offset = attr.ib(
        default=None,
        validator=attr.validators.optional(attr.validators.instance_of(float)),
    )
    power = attr.ib(
        default=Power.HIGH,
        validator=attr.validators.instance_of(Power),
        converter=Power.from_any,
    )
    rx_only = attr.ib(
        default=False,
        validator=attr.validators.instance_of(bool)
    )
    scanlist = attr.ib(default=None)
    code = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(str)))

    def __attrs_post_init__(self):
        type(self)._all_channels[self.name] = self

    @property
    def short_name(self):
        """Generate a unique short name for this channel"""
        attempt = 0
        maybe_this_channel = None
        while maybe_this_channel != self:
            maybe_short_name = dzcb.munge.channel_name(
                self.name,
                NAME_MAX - (1 if attempt else 0)
            ) + (str(attempt) if attempt else "")
            maybe_this_channel = type(self)._short_names.setdefault(maybe_short_name, self)
            attempt += 1
            if attempt > 9:
                raise RuntimeError("Cannot find a non-conflicting channel short name for {}".format(self))
        return maybe_short_name


@attr.s
class AnalogChannel(Channel):
    tone_encode = attr.ib(default=None)
    tone_decode = attr.ib(default=None)
    # configurable bandwidth for analog (technically should be enum)
    bandwidth = attr.ib(
        default=25,
        validator=attr.validators.instance_of(float),
        converter=float,
    )
    # configurable squelch for analog
    squelch = attr.ib(
        default=1,
        validator=attr.validators.instance_of(int),
        converter=int,
    )


@attr.s
class DigitalChannel(Channel):
    # fixed bandwidth for digital
    bandwidth = 12.5
    squelch = 0
    color_code = attr.ib(default=1)
    grouplist = attr.ib(default=None)
    talkgroup = attr.ib(default=None, validator=attr.validators.optional(attr.validators.instance_of(Talkgroup)))
    # a list of other static talkgroups, which form the basis of an RX/scan list
    static_talkgroups = attr.ib(factory=list)

    def from_talkgroups(self, talkgroups, order=None):
        """
        Return a channel per talkgroup based on this channel's settings
        """
        return [
            attr.evolve(
                self,
                name="{} {}".format(
                    tg.name_with_timeslot[: NAME_MAX - 4],
                    (self.code if self.code else self.name)[:3],
                ),
                talkgroup=tg,
                scanlist=self.short_name,
                static_talkgroups=[]
            ) for tg in dzcb.munge.ordered(
                seq=talkgroups,
                order=order or [],
                key=lambda tg: tg.name,
                log_sequence_name="{} static talkgroup list".format(self.short_name),
            )
        ]

    @property
    def zone_name(self):
        maybe_code = "{} ".format(self.code) if self.code else ""
        name = "{}{}".format(maybe_code, self.name)
        return name
        # shorten names for channel display
        # for old, new in self.name_replacements.items():
        #    name = name.replace(old, new)
        # return name


@attr.s
class GroupList:
    """
    A GroupList specifies a set of contacts that will be received on the same
    channel.
    """
    name = attr.ib(validator=attr.validators.instance_of(str))
    contacts = attr.ib(factory=list)


@attr.s
class ScanList:
    """
    A ScanList specifies a set of channels that can be sequentially scanned.
    """
    name = attr.ib(validator=attr.validators.instance_of(str))
    channels = attr.ib(factory=list, validator=attr.validators.deep_iterable(attr.validators.instance_of(Channel)))

    @classmethod
    def from_names(cls, name, channel_names):
        channels = []
        for cn in channel_names:
            # TODO: require method to be called with a set of available channels
            #       because the global list may contain channels that have already
            #       been pruned in the given codeplug
            channel = Channel._short_names.get(cn, Channel._all_channels.get(cn))
            if channel is None:
                logger.debug(
                    "ScanList {!r} references unknown channel {!r}, ignoring".format(
                        name, cn,
                    )
                )
                continue
            channels.append(channel)
        return cls(
            name=name,
            channels=channels
        )

    @classmethod
    def prune_missing_channels(cls, scanlists, channels):
        """
        Return a sequence of new ScanList objects containing only channels in `channels`
        """
        return [
            sl
            for sl in [
                attr.evolve(sl, channels=[ch for ch in sl.channels if ch in channels])
                for sl in scanlists
            ]
            if sl.channels
        ]

    @property
    def unique_channels(self):
        return list(self.channels)


@attr.s
class Zone:
    """
    A Zone groups channels together by name
    """

    name = attr.ib(validator=attr.validators.instance_of(str))
    channels_a = attr.ib(factory=list, validator=attr.validators.deep_iterable(attr.validators.instance_of(Channel)))
    channels_b = attr.ib(factory=list, validator=attr.validators.deep_iterable(attr.validators.instance_of(Channel)))

    @classmethod
    def prune_missing_channels(cls, zones, channels):
        """
        Return a sequence of new Zone objects containing only channels in `channels`
        """
        return [
            zn
            for zn in [
                attr.evolve(
                    zn,
                    channels_a=[ch for ch in zn.channels_a if ch in channels],
                    channels_b=[ch for ch in zn.channels_a if ch in channels],
                )
                for zn in zones
            ]
            if zn.unique_channels
        ]

    @property
    def unique_channels(self):
        channels = list(self.channels_a)
        for ch in self.channels_b:
            if ch not in channels:
                channels.append(ch)
        return channels


def uniquify_contacts(contacts):
    """
    Return a sequence of contacts with all duplicates removed.

    If any duplicate names are found without matching numbers, an exception is raised.
    """
    ctd = {}
    for ct in contacts:
        stored_ct = ctd.setdefault(ct.name, ct)
        if stored_ct.dmrid != ct.dmrid:
            raise RuntimeError("Two contacts named {} have different IDs: {} {}".format(ct.name, ct.dmrid, stored_ct.dmrid))
    return list(ctd.values())


def _seq_items_repr(s):
    return "<{} items>".format(len(s))


@attr.s
class Codeplug:
    contacts = attr.ib(factory=list, converter=uniquify_contacts, repr=_seq_items_repr)
    channels = attr.ib(factory=list, repr=_seq_items_repr)
    grouplists = attr.ib(factory=list, repr=_seq_items_repr)
    scanlists = attr.ib(factory=list, repr=_seq_items_repr)
    zones = attr.ib(factory=list, repr=_seq_items_repr)

    def __attrs_post_init__(self):
        # prune any channels which are not in a zone
        all_channels = []
        for z in self.zones:
            all_channels.extend(z.unique_channels)
        self.channels = [ch for ch in self.channels if ch in all_channels]

    def order_zones(self, zone_order=None, exclude_zones=None):
        if zone_order is None:
            zone_order = []
        if exclude_zones is None:
            exclude_zones = []

        zones = [
            zone
            for zone in dzcb.munge.ordered(
                seq=self.zones,
                order=zone_order,
                key=lambda z: z.name,
                log_sequence_name="zone list",
            )
            if zone.name not in exclude_zones
        ]
        return type(self)(
            contacts=list(self.contacts),
            channels=list(self.channels),
            grouplists=list(self.grouplists),
            scanlists=list(self.scanlists),
            zones=zones,
        )

    def order_grouplists(self, static_talkgroup_order=None, exclude_talkgroups=None):
        """
        Reorder the Talkgroups within each grouplist according to static_talkgroup_order.

        Exclude any talkgroups by name in exclude_talkgroups
        """
        if static_talkgroup_order is None:
            static_talkgroup_order = []
        if exclude_talkgroups is None:
            exclude_talkgroups = []

        grouplists = [
            attr.evolve(
                gl,
                contacts=dzcb.munge.ordered(
                    seq=[c for c in gl.contacts if c not in exclude_talkgroups],
                    order=static_talkgroup_order,
                    key=lambda c: c.name,
                    log_sequence_name="grouplist {}".format(gl.name),
                )
            )
            for gl in self.grouplists
        ]
        return type(self)(
            contacts=list(self.contacts),
            channels=list(self.channels),
            grouplists=grouplists,
            scanlists=list(self.scanlists),
            zones=list(self.zones),
        )

    def replace_scanlists(self, scanlist_dicts):
        """
        Return a new codeplug with additional scanlists.

        If the scanlist name appears in the codeplug, the entire scanlist
        is replaced by the new definition.

        If any channel names are not present in the codeplug, those channels
        will be ignored.

        :param scanlist_dicts: dict of scanlist_name -> [channel_name1, channel_name2, ...]
        """
        scanlists = {sl.name: sl for sl in self.scanlists}
        for sl_name, channels in scanlist_dicts.items():
            scanlists[sl_name] = ScanList.from_names(name=sl_name, channel_names=channels)

        return type(self)(
            contacts=list(self.contacts),
            channels=list(self.channels),
            grouplists=list(self.grouplists),
            scanlists=list(scanlists.values()),
            zones=list(self.zones),
        )

    def filter_frequency_range(self, *ranges):
        """
        :param ranges: tuple of (low, high) frequency to keep in the codeplug
        """
        def freq_in_range(freq):
            for low, high in ranges:
                if float(low) < freq < float(high):
                    return True
            return False

        channels = []
        channels_pruned = []

        for ch in self.channels:
            if freq_in_range(ch.frequency):
                channels.append(ch)
                continue
            # none of the ranges matched, so prune this channel
            channels_pruned.append(ch)
        if channels_pruned:
            logger.info(
                "filter_frequency_range: Excluding %s channels with frequency out of range: %s",
                len(channels_pruned),
                ranges,
            )

        return type(self)(
            contacts=list(self.contacts),
            channels=channels,
            grouplists=list(self.grouplists),
            scanlists=ScanList.prune_missing_channels(self.scanlists, channels),
            zones=Zone.prune_missing_channels(self.zones, channels),
        )

    def expand_static_talkgroups(self, static_talkgroup_order=None):
        """
        This function replaces channels with static_talkgroups by a zone
        containing a channel per talkgroup in static_talkgroups.

        :return: new Codeplug with additional channels and zones for each
            DigitalChannel with static_talkgroups
        """
        if static_talkgroup_order is None:
            static_talkgroup_order = []
        zones = list(self.zones)
        channels = []
        exp_scanlists = []
        for ch in self.channels:
            if not isinstance(ch, DigitalChannel) or not ch.static_talkgroups:
                channels.append(ch)
                continue
            zone_channels = ch.from_talkgroups(ch.static_talkgroups, order=static_talkgroup_order)
            zscanlist = ScanList(
                name=ch.short_name,
                channels=zone_channels,
            )
            exp_scanlists.append(zscanlist)
            zones.append(
                Zone(
                    name=ch.short_name,
                    channels_a=zone_channels,
                    channels_b=zone_channels,
                )
            )
            channels.extend(zone_channels)

        return type(self)(
            contacts=list(self.contacts),
            channels=channels,
            grouplists=list(self.grouplists),
            # Don't reference channels that no longer exist
            scanlists=ScanList.prune_missing_channels(self.scanlists, channels) + exp_scanlists,
            zones=Zone.prune_missing_channels(zones, channels),
        )



"""
dzcb.pnwdigital - cache data from PNWDigital.net
"""
import argparse
import logging
from pathlib import Path
import os
import tempfile
from zipfile import ZipFile

import requests

logger = logging.getLogger(__name__)

PNWDIGITAL_REPEATERS = "http://www.pnwdigital.net/files/acb/PNW_Digital_Repeaters.zip"
REPEATER_FILENAME = "Digital-Repeaters__PNWDigital.csv"
TALKGROUPS_FILENAME = "Talkgroups__PNWDigital.csv"


def cache_repeaters(output_dir):
    resp = requests.get(PNWDIGITAL_REPEATERS)
    resp.raise_for_status()
    with tempfile.TemporaryFile() as tf:
        tf.write(resp.content)
        zf = ZipFile(tf, "r")
        names = zf.namelist()
        zip_repeater_filename = [
            n for n in names if n.startswith("Digital-Repeaters__PNW-all")
        ]
        if len(zip_repeater_filename) > 1:
            raise RuntimeError(
                "Multiple Digital-Repeaters found in the zip: {}".format(
                    zip_repeater_filename
                )
            )
        output_repeaters = Path(output_dir) / REPEATER_FILENAME
        output_repeaters.write_bytes(
            zf.read(zip_repeater_filename[0])
        )
        logger.info("Cache PNWDigital k7abd zones to '%s'", output_repeaters)

        zip_talkgroups_filename = [
            n for n in names if n.startswith("Talkgroups__PNW-all")
        ]
        if len(zip_talkgroups_filename) > 1:
            raise RuntimeError(
                "Multiple Talkgroups found in the zip: {}".format(
                    zip_talkgroups_filename
                )
            )
        output_talkgroups = Path(output_dir) / TALKGROUPS_FILENAME
        output_talkgroups.write_bytes(
            zf.read(zip_talkgroups_filename[0])
        )
        logger.info("Cache PNWDigital k7abd talkgroups to '%s'", output_talkgroups)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("output_dir")
    args = parser.parse_args()
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    cache_repeaters(args.output_dir)



"""
Write series of CSV files acceptable for import into Anytone CPS tool

Supported CPS versions

    578: 1.11
    868: 1.39
    878: 1.21

"""
import csv
import enum
import logging
from pathlib import Path

from dzcb import AMATEUR_220, COMMERCIAL_UHF, COMMERCIAL_VHF
from dzcb.model import AnalogChannel, DigitalChannel

logger = logging.getLogger(__name__)

NAME_MAX = 16
SCANLIST_MAX = 50
OFF = "Off"
ON = "On"
NONE = "None"

value_replacements = {
    None: NONE,
    False: OFF,
    True: ON,
}

format_frequency = "{:.5f}".format
format_channel_type = {
    AnalogChannel: "A-Analog",
    DigitalChannel: "D-Digital",
}.get


def format_member_list(members, list_name, expand_members=False):
    d = {list_name: "|".join(m.short_name for m in members)}
    if expand_members:
        d.update(
            {
                list_name
                + " RX Frequency": "|".join(
                    format_frequency(m.frequency) for m in members
                ),
                list_name
                + " TX Frequency": "|".join(
                    format_frequency(m.frequency + m.offset) for m in members
                ),
            }
        )
    return d


def replace_field_names(d, model):
    for orig, repl in model.get("replace_field_names", {}).items():
        if orig in d:
            d[repl] = d[orig]
            del d[orig]
    return d


def remove_fields(d, model):
    for f in model.get("remove_fields", []):
        if f in d:
            del d[f]
    return d


class DMR_MODE(enum.Enum):
    SIMPLEX = 0
    REPEATER = 1
    DUAL_SLOT = 2

    @classmethod
    def value_from(cls, channel):
        if abs(channel.offset) > 0:
            return cls.REPEATER.value
        return cls.SIMPLEX.value


class TXPermit(enum.Enum):
    ALWAYS = "Always"
    CHANNELFREE = "ChannelFree"
    SAMECOLOR = "Same Color Code"
    DIFFERENTCOLOR = "Different Color Code"

    @classmethod
    def value_from(cls, channel):
        if abs(channel.offset) > 0:
            return cls.SAMECOLOR.value
        return cls.ALWAYS.value


# 578/868/878 Common Talkgroups.CSV format
talkgroup_fields = ("No.", "Radio ID", "Name", "Call Type", "Call Alert")
talkgroup_filename = "TalkGroups.CSV"
talkgroup_filename_578_1_11 = "ContactTalkGroups.CSV"

channel_fields_578_1_11 = {
    "No.": None,
    "Channel Name": None,
    "Receive Frequency": None,
    "Transmit Frequency": None,
    "Channel Type": None,
    "Transmit Power": "High",
    "Band Width": "25K",
    "CTCSS/DCS Decode": OFF,
    "CTCSS/DCS Encode": OFF,
    "Contact": "",
    "Contact Call Type": "Group Call",
    "Contact TG/DMR ID": "0",
    "Radio ID": "",
    "Busy Lock/TX Permit": "Always",
    "Squelch Mode": "Carrier",
    "Optional Signal": OFF,
    "DTMF ID": "1",
    "2Tone ID": "1",
    "5Tone ID": "1",
    "PTT ID": OFF,
    "Color Code": "1",
    "Slot": "1",
    "Scan List": NONE,
    "Receive Group List": NONE,
    "PTT Prohibit": OFF,
    "Reverse": OFF,
    "TDMA": OFF,
    "TDMA Adaptive": OFF,
    "AES Digital Encryption": "Normal Encryption",
    "Digital Encryption": OFF,
    "Call Confirmation": OFF,
    "Talk Around(Simplex)": OFF,
    "Work Alone": OFF,
    "Custom CTCSS": "251.1",
    "2TONE Decode": "0",
    "Ranging": OFF,
    "Simplex": OFF,
    "Digi APRS RX": OFF,
    "Analog APRS PTT Mode": OFF,
    "Digital APRS PTT Mode": OFF,
    "APRS Report Type": OFF,
    "Digital APRS Report Channel": "1",
    "Correct Frequency[Hz]": "0",
    "SMS Confirmation": OFF,
    "Exclude Channel From Roaming": "0",
    "DMR MODE": "0",
    "DataACK Disable": "0",
    "R5toneBot": "0",
    "R5ToneEot": "0",
}
channel_fields_868_1_39 = {
    "No.": None,
    "Channel Name": None,
    "Receive Frequency": None,
    "Transmit Frequency": None,
    "Channel Type": None,
    "Transmit Power": "High",
    "Band Width": "25K",
    "CTCSS/DCS Decode": OFF,
    "CTCSS/DCS Encode": OFF,
    "Contact": "",
    "Contact Call Type": "Group Call",
    "Radio ID": "",
    "Busy Lock/TX Permit": "Always",
    "Squelch Mode": "Carrier",
    "Optional Signal": OFF,
    "DTMF ID": "1",
    "2Tone ID": "1",
    "5Tone ID": "1",
    "PTT ID": OFF,
    "Color Code": "1",
    "Slot": "1",
    "CH Scan List": NONE,
    "Receive Group List": NONE,
    "TX Prohibit": OFF,
    "Reverse": OFF,
    "Simplex TDMA": OFF,
    "TDMA Adaptive": OFF,
    "Encryption Type": "Normal Encryption",
    "Digital Encryption": OFF,
    "Call Confirmation": OFF,
    "Talk Around": OFF,
    "Work Alone": OFF,
    "Custom CTCSS": "251.1",
    "2TONE Decode": "0",
    "Ranging": OFF,
    "Through Mode": OFF,
    "APRS Report": OFF,
    "APRS Report Channel": "1",
}
channel_fields_878_1_21 = {
    "No.": None,
    "Channel Name": None,
    "Receive Frequency": None,
    "Transmit Frequency": None,
    "Channel Type": None,
    "Transmit Power": "High",
    "Band Width": "25K",
    "CTCSS/DCS Decode": OFF,
    "CTCSS/DCS Encode": OFF,
    "Contact": "",
    "Contact Call Type": "Group Call",
    "Contact TG/DMR ID": "0",
    "Radio ID": "",
    "Busy Lock/TX Permit": "Always",
    "Squelch Mode": "Carrier",
    "Optional Signal": OFF,
    "DTMF ID": "1",
    "2Tone ID": "1",
    "5Tone ID": "1",
    "PTT ID": OFF,
    "Color Code": "1",
    "Slot": "1",
    "Scan List": NONE,
    "Receive Group List": NONE,
    "PTT Prohibit": OFF,
    "Reverse": OFF,
    "Simplex TDMA": OFF,
    "Slot Suit": OFF,
    "AES Digital Encryption": "Normal Encryption",
    "Digital Encryption": OFF,
    "Call Confirmation": OFF,
    "Talk Around(Simplex)": OFF,
    "Work Alone": OFF,
    "Custom CTCSS": "251.1",
    "2TONE Decode": "0",
    "Ranging": OFF,
    "Through Mode": OFF,
    "Digi APRS RX": OFF,
    "Analog APRS PTT Mode": OFF,
    "Digital APRS PTT Mode": OFF,
    "APRS Report Type": OFF,
    "Digital APRS Report Channel": "1",
    "Correct Frequency[Hz]": "0",
    "SMS Confirmation": OFF,
    "Exclude Channel From Roaming": "0",
    "DMR MODE": "0",
    "DataACK Disable": "0",
    "R5toneBot": "0",
    "R5ToneEot": "0",
}
channel_filename = "Channel.CSV"

scanlist_fields_578_1_11 = {
    "No.": None,
    "Scan List Name": None,
    "Scan Channel Member": None,
    "Scan Channel Member RX Frequency": None,
    "Scan Channel Member TX Frequency": None,
    "Scan Mode": OFF,
    "Priority Channel Select": "Priority Channel Select1",
    "Priority Channel 1": "Current Channel",
    "Priority Channel 1 RX Frequency": "",
    "Priority Channel 1 TX Frequency": "",
    "Priority Channel 2": OFF,
    "Priority Channel 2 RX Frequency": "",
    "Priority Channel 2 TX Frequency": "",
    "Revert Channel": "Selected",
    "Look Back Time A[s]": "2.0",
    "Look Back Time B[s]": "3.0",
    "Dropout Delay Time[s]": "3.1",
    "Dwell Time[s]": "3.1",
}
scanlist_fields_868_1_39 = {
    "No.": None,
    "Scan List Name": None,
    "Scan Channel Member": None,
    "Scan Mode": OFF,
    "Priority Channel Select": "Priority Channel Select1",
    "Priority Channel 1": "Current Channel",
    "Priority Channel 2": OFF,
    "Revert Channel": "Selected",
    "Look Back Time A[s]": "2.0",
    "Look Back Time B[s]": "3.0",
    "Dropout Delay Time[s]": "3.1",
    "Dwell Time[s]": "3.1",
}
scanlist_filename = "ScanList.CSV"

zone_fields_578_1_11 = {
    "No.": None,
    "Zone Name": None,
    "Zone Channel Member": None,
    "Zone Channel Member RX Frequency": None,
    "Zone Channel Member TX Frequency": None,
    "A Channel": None,
    "A Channel RX Frequency": None,
    "A Channel TX Frequency": None,
    "B Channel": None,
    "B Channel RX Frequency": None,
    "B Channel TX Frequency": None,
}
zone_fields_868_1_39 = {
    "No.": None,
    "Zone Name": None,
    "Zone Channel Member": None,
    "A Channel": None,
    "B Channel": None,
}
zone_filename = "Zone.CSV"

SUPPORTED_RADIOS = {
    "578_1_11": dict(
        version="1.11",
        expand_members=True,
        frequency_range=(COMMERCIAL_VHF, AMATEUR_220, COMMERCIAL_UHF),
        channel=channel_fields_578_1_11,
        channel_filename=channel_filename,
        scanlist=scanlist_fields_578_1_11,
        scanlist_filename=scanlist_filename,
        zone=zone_fields_578_1_11,
        zone_filename=zone_filename,
        talkgroup=talkgroup_fields,
        talkgroup_filename=talkgroup_filename_578_1_11,
        replace_field_names={
            "Through Mode": "Simplex",
        },
        remove_fields=[],
    ),
    "868_1_39": dict(
        version="1.39",
        expand_members=False,
        frequency_range=(COMMERCIAL_VHF, COMMERCIAL_UHF),
        channel=channel_fields_868_1_39,
        channel_filename=channel_filename,
        scanlist=scanlist_fields_868_1_39,
        scanlist_filename=scanlist_filename,
        zone=zone_fields_868_1_39,
        zone_filename=zone_filename,
        talkgroup=talkgroup_fields,
        talkgroup_filename=talkgroup_filename,
        replace_field_names={
            "PTT Prohibit": "TX Prohibit",
            "Scan List": "CH Scan List",
        },
        remove_fields=["Contact TG/DMR ID", "DMR MODE"],
    ),
    "878_1_21": dict(
        version="1.21",
        expand_members=True,
        frequency_range=(COMMERCIAL_VHF, COMMERCIAL_UHF),
        channel=channel_fields_878_1_21,
        channel_filename=channel_filename,
        scanlist=scanlist_fields_578_1_11,
        scanlist_filename=scanlist_filename,
        zone=zone_fields_578_1_11,
        zone_filename=zone_filename,
        talkgroup=talkgroup_fields,
        talkgroup_filename=talkgroup_filename,
        replace_field_names={},
        remove_fields=[],
    ),
}


def Talkgroup_to_dict(index, talkgroup):
    return {
        "No.": str(index + 1),
        "Radio ID": str(talkgroup.dmrid),
        "Name": talkgroup.name,
        "Call Type": str(talkgroup.kind) + " Call",
        "Call Alert": NONE,
    }


def Talkgroup_to_channel_detail(talkgroup):
    return {
        "Contact": talkgroup.name,
        "Contact Call Type": str(talkgroup.kind) + " Call",
        "Contact TG/DMR ID": talkgroup.dmrid,
        "Slot": str(talkgroup.timeslot),
    }


def AnalogChannel_to_dict(channel):
    return {
        "CTCSS/DCS Decode": channel.tone_decode or OFF,
        "CTCSS/DCS Encode": channel.tone_encode or OFF,
        "Squelch Mode": "CTCSS/DCS" if channel.tone_decode else "Carrier",
        "Busy Lock/TX Permit": OFF,
    }


def DigitalChannel_to_dict(channel):
    d = {
        "Color Code": str(channel.color_code),
        "Scan List": channel.scanlist,
        "Busy Lock/TX Permit": TXPermit.value_from(channel),
        "DMR MODE": DMR_MODE.value_from(channel),
        # On the 578 and 878, DMR MODE = "Simplex" (0) channels
        # also have "Simplex=On" and "Through Mode=On" in the
        # exported file. Neither targeted CPS version exposes
        # this setting in the UI. But the 578 1.11 CPS will show
        # DMR MODE "Repeater" unless "Simplex=On"
        # Set it ON for simplex channels, and it will have zero
        # effect because it only makes a difference on split channels
        "Through Mode": OFF if abs(channel.offset) > 0 else ON,
        # TODO: Support group list
    }
    if channel.talkgroup:
        d.update(Talkgroup_to_channel_detail(channel.talkgroup))
    return d


def Channel_to_dict(index, channel):
    d = {
        "No.": str(index + 1),
        "Channel Name": channel.short_name,
        "Receive Frequency": format_frequency(channel.frequency),
        "Transmit Frequency": format_frequency(channel.frequency + channel.offset),
        "Channel Type": format_channel_type(type(channel)),
        "Transmit Power": str(channel.power),
        "Band Width": "25K" if channel.bandwidth > 19 else "12.5K",
        "PTT Prohibit": value_replacements[channel.rx_only],
    }
    if isinstance(channel, AnalogChannel):
        d.update(AnalogChannel_to_dict(channel))
    else:
        d.update(DigitalChannel_to_dict(channel))
    return d


def Zone_to_dict(index, zone, expand_members):
    d = {
        "No.": str(index + 1),
        "Zone Name": zone.name,
    }
    d.update(
        format_member_list(
            members=zone.unique_channels,
            list_name="Zone Channel Member",
            expand_members=expand_members,
        )
    )
    for list_name in ("A Channel", "B Channel"):
        d.update(
            format_member_list(
                members=(zone.unique_channels[0],),
                list_name=list_name,
                expand_members=expand_members,
            )
        )
    return d


def ScanList_to_dict(index, scanlist, expand_members):
    d = {
        "No.": str(index + 1),
        "Scan List Name": scanlist.name,
    }
    d.update(
        format_member_list(
            members=scanlist.unique_channels[:SCANLIST_MAX],
            list_name="Scan Channel Member",
            expand_members=expand_members,
        )
    )
    return d


def Codeplug_to_anytone_csv(cp, output_dir, models=None):
    if models is None:
        models = tuple(SUPPORTED_RADIOS.keys())
    for model_id in models:
        radio_dir = Path(output_dir) / model_id
        radio_dir.mkdir(parents=True, exist_ok=True)
        model = SUPPORTED_RADIOS[model_id]
        # filter down to supported frequency ranges
        mcp = cp.filter_frequency_range(*model["frequency_range"])
        with (radio_dir / model["talkgroup_filename"]).open("w", newline="") as f:
            csvw = csv.DictWriter(f, model["talkgroup"])
            csvw.writeheader()
            for ix, tg in enumerate(mcp.contacts):
                csvw.writerow(Talkgroup_to_dict(ix, tg))
        with (radio_dir / model["channel_filename"]).open("w", newline="") as f:
            csvw = csv.DictWriter(f, tuple(model["channel"].keys()))
            csvw.writeheader()
            for ix, channel in enumerate(mcp.channels):
                d = model["channel"].copy()
                d.update(Channel_to_dict(ix, channel))
                csvw.writerow(replace_field_names(remove_fields(d, model), model))
        with (radio_dir / model["zone_filename"]).open("w", newline="") as f:
            csvw = csv.DictWriter(f, tuple(model["zone"].keys()))
            csvw.writeheader()
            for ix, zone in enumerate(mcp.zones):
                csvw.writerow(Zone_to_dict(ix, zone, model["expand_members"]))
        with (radio_dir / model["scanlist_filename"]).open("w", newline="") as f:
            csvw = csv.DictWriter(f, tuple(model["scanlist"].keys()))
            csvw.writeheader()
            for ix, sl in enumerate(mcp.scanlists):
                d = model["scanlist"].copy()
                d.update(ScanList_to_dict(ix, sl, model["expand_members"]))
                csvw.writerow(d)
        logger.info("Wrote Anytone %s CSV files to '%s'", model_id, radio_dir)



"""
dzcb.munge - replacements, filters, and modifications of the data
"""
import re
import warnings

# These are used when generating channel names
Talkgroup_Channel_name_replacements = {
    "Audio Test": "A.Test",
    "California": "CA",
    "English": "Eng",
    "Hawaii": "HI",
    "Idaho": "ID",
    "Montana": "MT",
    "Oregon": "OR",
    "Utah": "UT",
    "Washington": "WA",
    "Worldwide": "WW",
}


def channel_name(ch_name, max_length):
    # Replace Long strings with shorter ones
    replacements = Talkgroup_Channel_name_replacements.copy()
    for find, repl in replacements.items():
        ch_name = ch_name.replace(find, repl)

    # Truncate the channel name (try to preserve the tail  characters
    # which are typically TG# and 3-digit Code)
    tail_code = re.search(r"[12]?\s[A-Z]+$", ch_name)
    if len(ch_name) > max_length and tail_code:
        n_tail = len(tail_code.group())
        if max_length > n_tail + 1:
            n_trunc = len(ch_name) - max_length
            ch_name = ch_name[: -n_trunc - n_tail] + ch_name[-n_tail:]

    return ch_name[:max_length]


def contact_name(contact_name):
    # PNWDigital source annoyinging appends "-2" to the TAC talkgroup names
    # strip off the suffix because not all systems have the TACs on TS 2
    # Radios such as the 868 can only map a given DMR ID to a single name
    # so the "-2" suffix is confusing
    if contact_name.startswith("TAC") and contact_name.endswith("-2"):
        return contact_name[:-2]
    return contact_name


def zone_name(zone_name, max_length):
    return zone_name[:max_length]


class MissingItemsWarning(UserWarning):
    """Raised when items in order list do not appear in the sequence"""

    def __init__(self, missing_items, sequence, log_sequence_name=None):
        self.sequence = sequence
        self.missing_items = missing_items
        if log_sequence_name is None:
            log_sequence_name = "the sequence to be ordered"
        self.log_sequence_name = log_sequence_name

    def __str__(self):
        return "Items were not in {}: {}".format(
            self.log_sequence_name, self.missing_items
        )


def ordered(seq, order, key=None, log_sequence_name=None):
    """
    If `log_sequence_name` is specified, use that text instead of
    "the sequence to be ordered" when emitting a log message about
    missing items.
    """
    NotFound = object()
    head = [NotFound] * len(order)
    tail = []
    for item in seq:
        k = key(item) if key else item
        try:
            head[order.index(k)] = item
        except ValueError:
            tail.append(item)
    if NotFound in head:
        warnings.warn(
            MissingItemsWarning(
                missing_items=[
                    order[ix] for ix, item in enumerate(head) if item is NotFound
                ],
                sequence=seq,
                log_sequence_name=log_sequence_name,
            ),
            stacklevel=2,
        )
        head = [item for item in head if item is not NotFound]
    return head + tail



"""
dzcb: DMR Zone Channel Builder

Fetch - fetch input files from local directories or urls
Assemble - combine information from multiple sources
Filter - rename, exclude, reorder zones, and talkgroups
Format - output to common export formats
"""
import enum

import appdirs
from pkg_resources import get_distribution

__version__ = get_distribution(__name__).version

COMMERCIAL_VHF = (136.0, 174.0)
COMMERCIAL_UHF = (400.0, 480.0)
AMATEUR_220 = (222.0, 225.0)


class AmateurBands(enum.Enum):
    B_10m = (28.0, 29.7)
    B_6m = (50.0, 54.0)
    B_2m = (144.0, 148.0)
    B_125cm = (222.0, 225.0)
    B_70cm = (420.0, 450.0)
    B_33cm = (902.0, 928.0)
    B_23cm = (1240.0, 1300.0)

    @classmethod
    def get_normalized(cls, input):
        if str(input).startswith("B_"):
            return getattr(cls, str(input))
        all_bands = cls.__members__
        maybe_band = "B_" + str(input)
        if maybe_band in all_bands:
            return getattr(cls, maybe_band)
        if str(input).lower() in ("1.25m", "220"):
            return cls.B_125cm
        try:
            input = float(input)
        except ValueError:
            pass
        if isinstance(input, (int, float)):
            # return the band that the frequency is from
            for band in all_bands.values():
                f_low, f_high = band.value
                if f_low <= input <= f_high:
                    return band
        raise ValueError(
            "Unknown amateur band/frequency: {!r}\nChoose from {!r}".format(
                input, list(all_bands.values())
            ),
        )


appdir = appdirs.AppDirs("dzcb", "mycodeplug", version=__version__)



"""
dzcb.exceptions - errors that may be raised
"""


class InvalidDmrID(ValueError):
    pass


class DuplicateDmrID(ValueError):
    def __init__(self, msg, existing_contact):
        super().__init__(msg)
        self.existing_contact = existing_contact



"""
dzcb.repeaterbook - export CSV from Repeaterbook, convert to K7ABD format
"""
import argparse
import csv
import hashlib
import json
import logging
from pathlib import Path
import os
import time

import geopy.distance
import requests

from . import appdir, AmateurBands

logger = logging.getLogger(__name__)

# ?country=United%20States&state=Washington&state=Oregon&state=Idaho&state=California"
REPEATERBOOK_API = "https://www.repeaterbook.com/api/export.php"
REPEATERBOOK_API_DELAY = 30
REPEATERBOOK_LAST_FETCH = 0

# XXX: Repeaterbook API returns 3500 records per request,
#      so limit the data to the area of interest. Eventually
#      this list will be passed by JSON or CSV
REPEATERBOOK_DEFAULT_STATES = ("Washington","Oregon","Idaho","California","British Columbia")
REPEATERBOOK_CACHE_MAX_AGE = 3600 * 12.1  # 12 hours (and some change)
CSV_ZONE_NAME = "Zone Name"
CSV_LAT = "Lat"
CSV_LONG = "Long"
CSV_DISTANCE = "Distance"
CSV_UNIT = "Unit"
CSV_BAND = "Band(2m;1.25m;70cm)"


def cached_json(url, max_age=REPEATERBOOK_CACHE_MAX_AGE):
    md5urlhash = hashlib.md5(url.encode("utf-8")).hexdigest()
    cachedir = Path(appdir.user_cache_dir)
    filepath = cachedir / "repeaters_{}.json".format(md5urlhash)
    if not filepath.exists() or filepath.stat().st_mtime < time.time() - max_age:
        # cache is expired, need to refetch
        cachedir.mkdir(parents=True, exist_ok=True)
        # don't make requests too often
        global REPEATERBOOK_LAST_FETCH
        global_last_fetched = time.time() - REPEATERBOOK_LAST_FETCH
        if global_last_fetched < REPEATERBOOK_API_DELAY:
            time.sleep(REPEATERBOOK_API_DELAY - global_last_fetched)
        resp = requests.get(url)
        REPEATERBOOK_LAST_FETCH = time.time()
        filepath.write_bytes(resp.content)
    return filepath


def iter_cached_repeaters(states=None, max_age=REPEATERBOOK_CACHE_MAX_AGE):
    if states is None:
        states = REPEATERBOOK_DEFAULT_STATES
    for state in states:
        url = "".join(
            (
                REPEATERBOOK_API,
                "?state={}".format(state),
            )
        )
        cached_json_file = cached_json(url, max_age=max_age)
        with open(cached_json_file, "r") as f:
            rb_api_resp = json.load(f)
            logger.info(
                "Load cached Repeaterbook data for %s: %s records (%s)",
                state,
                len(rb_api_resp["results"]),
                cached_json_file,
            )
        for repeater in rb_api_resp["results"]:
            yield repeater


def proximity_zones(proximity_zones_csv):
    csvr = csv.DictReader(
        proximity_zones_csv,
    )
    for zone in csvr:
        name = zone.pop(CSV_ZONE_NAME)
        slug = name.replace(" ", "-").replace(",", "")
        yield (name, slug, zone)


def matches_criteria(repeater, criteria):
    for field, value in criteria.items():
        # XXX: maybe need a regex match here...
        if str(value).lower() != str(repeater.get(field, "")).lower():
            return False
    return True


def filter_repeaters(repeaters, zone):
    zone = zone.copy()
    radius = zone.pop(CSV_DISTANCE)
    dunit = zone.pop(CSV_UNIT)
    if radius:
        radius = float(radius)
        poi_coords = (float(zone.pop(CSV_LAT)), float(zone.pop(CSV_LONG)))
    else:
        distance = 0
    bands = [
        AmateurBands.get_normalized(b)
        for b in zone.pop(CSV_BAND).strip().split(";")
        if b
    ]
    matching = []
    # Find matching repeaters
    for r in repeaters:
        if not r:
            continue
        if radius:
            # repeater must be within radius of poi
            repeater_coords = (r["Lat"], r["Long"])
            distance = geopy.distance.distance(poi_coords, repeater_coords)
            if getattr(distance, dunit) > radius:
                continue
        if bands:
            # repeater frequency must be in the given bands
            if AmateurBands.get_normalized(r["Frequency"]) not in bands:
                continue

        # remaining fields in the zone list are criteria to satisfy
        if matches_criteria(r, zone):
            matching.append((distance, r))
    return [r for _, r in sorted(matching, key=lambda x: x[0])]


def zones_to_k7abd(input_csv, output_dir, states=None):
    repeaters = list(iter_cached_repeaters(states=states))
    for name, slug, zone in proximity_zones(input_csv):
        out_file = Path(output_dir) / "Analog__{}.csv".format(slug)
        total_channels = 0
        with open(out_file, "w", newline="") as out:
            csvw = csv.DictWriter(
                out,
                fieldnames=[
                    "Zone",
                    "Channel Name",
                    "Bandwidth",
                    "Power",
                    "RX Freq",
                    "TX Freq",
                    "CTCSS Decode",
                    "CTCSS Encode",
                    "TX Prohibit",
                ],
            )
            csvw.writeheader()
            for repeater in filter_repeaters(repeaters, zone):
                csvw.writerow(
                    {
                        "Zone": name,
                        "Channel Name": "{} {} {}".format(
                            repeater["Callsign"],
                            repeater["Nearest City"],
                            repeater["Landmark"],
                        ).strip(),
                        "Bandwidth": "25K",
                        "Power": "High",
                        "RX Freq": repeater["Frequency"],
                        "TX Freq": repeater["Input Freq"] or repeater["Frequency"],
                        "CTCSS Decode": repeater["TSQ"] or "Off",
                        "CTCSS Encode": repeater["PL"] or "Off",
                        "TX Prohibit": "Off",
                    }
                )
                total_channels += 1
        logger.debug("Generate '%s' k7abd zones (%s channels) to '%s'", name, total_channels, out_file)


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument("proximity_csv_file", type=argparse.FileType("r"))
    parser.add_argument("output_dir")
    args = parser.parse_args()
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    zones_to_k7abd(args.proximity_csv_file, args.output_dir)



"""
dzcb.k7abd - read k7abd anytone-codeplug-builder CSV files

These files are provided by PNWDigital and generated from repeaterbook CSV,
and fully describe how to build a complex code plug.

A k7abd is given as a directory or a .zip file and contains files of the
following format.

  * Analog__name-%Y-%m-%d.csv
    * Analog simplex and repeater frequencies. The "name" in the filename is
      not used. Only the Zone in the CSV makes it to the codeplug.
  * Digital-Others__name-%Y-%m-%d.csv
    * Digital simplex channels with 1:1 talkgroup mapping
  * Digital-Repeaters__name-%Y-%m-%d.csv
    * Digital repeater network provides talkgroup matrix
    * 1 channel is created per talkgroup per repeater. In this configuration
      the Zone name is the repeater name/code
  * Talkgroups__name-%Y-%m-%d.csv
    * Provides talkgroup name/number mapping
    * "name" in the filename should match with an associated Digital-Repeaters CSV
"""
import csv
import logging
from pathlib import Path

import attr

import dzcb.exceptions
from dzcb.model import AnalogChannel, Codeplug, Contact, DigitalChannel, GroupList, ScanList, Talkgroup, Zone


logger = logging.getLogger(__name__)


def Talkgroups_map_from_csv(talkgroups_csv):
    talkgroups_by_name = {}
    for tg_name, tg_id in csv.reader(talkgroups_csv):
        try:
            talkgroups_by_name.setdefault(
                tg_name,
                Contact(
                    name=tg_name,
                    dmrid=tg_id,
                ),
            )
        except dzcb.exceptions.DuplicateDmrID as dup:
            talkgroups_by_name.setdefault(
                tg_name,
                dup.existing_contact,
            )
            logger.debug(
                "Mapping %s (%s) to existing contact %s (%s)",
                tg_name,
                tg_id,
                dup.existing_contact.name,
                dup.existing_contact.dmrid,
            )
    return talkgroups_by_name


def Codeplug_from_zone_dicts(zone_dicts):
    """
    :param zone_dicts: dict of ZoneName -> [DigitalChannel, AnalogChannel, etc... ]
    """
    contacts = set()
    channels = list()
    grouplists = list()
    scanlists = list()
    zones = list()

    def update_static_talkgroups(ch):
        contacts.update(ch.static_talkgroups)
        grouplist = GroupList(
            name="{} TGS".format(ch.code),
            contacts=ch.static_talkgroups,
        )
        grouplists.append(grouplist)
        return attr.evolve(ch, grouplist=grouplist)

    all_channels = {}
    for zname, zchannels in zone_dicts.items():
        updated_channels = []
        for ch in zchannels:
            if isinstance(ch, DigitalChannel):
                if ch.static_talkgroups:
                    ch = update_static_talkgroups(ch)
                if ch.talkgroup:
                    contacts.add(ch.talkgroup)
            if ch.scanlist is None:
                ch = attr.evolve(ch, scanlist=zname)
            if ch.short_name not in all_channels:
                all_channels[ch.short_name] = ch
            updated_channels.append(ch)
        zscanlist = ScanList(
            name=zname,
            channels=updated_channels,
        )
        scanlists.append(zscanlist)
        zones.append(
            Zone(
                name=zname,
                channels_a=updated_channels,
                channels_b=updated_channels,
            )
        )
    channels.extend(all_channels.values())
    return Codeplug(
        contacts=sorted(list(contacts), key=lambda c: c.name),
        channels=channels,
        grouplists=grouplists,
        scanlists=sorted(list(scanlists), key=lambda s: s.name),
        zones=sorted(zones, key=lambda z: z.name),
    )


def Analog_from_csv(analog_repeaters_csv):
    zones = {}
    csvr = csv.DictReader(analog_repeaters_csv)
    for r in csvr:
        zname = r["Zone"]
        zname, found, code = zname.partition(";")
        name = r["Channel Name"]
        frequency = float(r["RX Freq"])
        offset = round(float(r["TX Freq"]) - frequency, 1)
        power = r["Power"]
        bandwidth = r["Bandwidth"].rstrip("K")
        tone_encode = r["CTCSS Decode"] if r["CTCSS Decode"] != "Off" else None
        tone_decode = r["CTCSS Encode"] if r["CTCSS Encode"] != "Off" else None
        zones.setdefault(zname, []).append(
            AnalogChannel(
                name=name,
                code=code or None,
                frequency=frequency,
                offset=offset,
                tone_encode=tone_encode,
                tone_decode=tone_decode,
                power=power,
                bandwidth=bandwidth,
            )
        )
    return zones


def DigitalRepeaters_from_k7abd_csv(digital_repeaters_csv, talkgroups_by_name):
    """
    read a talkgroup matrix and yield DigitalChannel

    :param digital_repeaters_csv: iterable of CSV lines: ... see code ;]
    :param talkgroups_by_name: map of tg_name -> Talkgroup
    :return: iterable of DigitalChannel with static_talkgroups ready to be expanded
        and converted into group/scan lists.
    """
    csvr = csv.DictReader(digital_repeaters_csv)
    for r in csvr:
        _ = r.pop("Comment", None)
        zname, found, code = r.pop("Zone Name").partition(";")
        frequency = float(r.pop("RX Freq"))
        if not frequency:
            logger.info("%s: Excluding repeater, %s with no frequency", digital_repeaters_csv, zname)
            continue
        offset = round(float(r.pop("TX Freq")) - frequency, 1)
        color_code = r.pop("Color Code")
        power = r.pop("Power")
        talkgroups = []
        for tg_name, timeslot in r.items():
            if timeslot.strip() == "-":
                continue
            try:
                talkgroups.append(
                    Talkgroup.from_contact(
                        talkgroups_by_name[tg_name], timeslot,
                    )
                )
            except ValueError:
                logger.info("%s: Ignoring ValueError from %s:%s", digital_repeaters_csv, tg_name, timeslot)
        repeater = DigitalChannel(
            name=zname,
            code=code or None,
            frequency=frequency,
            offset=offset,
            color_code=color_code,
            power=power,
            static_talkgroups=sorted(talkgroups, key=lambda tg: tg.name),
        )
        yield repeater


def DigitalChannels_from_k7abd_csv(digital_others_csv, talkgroups_by_name):
    """
    read a Digital-Others files and yield DigitalChannel

    :param digital_others_csv: iterable of CSV lines: ... see code ;]
    :param talkgroups_by_name: map of tg_name -> Talkgroup
    :return: dict of zone_name -> tuple of DigitalChannel (with talkgroup set)
    """
    zones = {}
    csvr = csv.DictReader(digital_others_csv)
    for r in csvr:
        _ = r.pop("Comment", None)
        zname, found, code = r.pop("Zone Name", r.pop("Zone")).partition(";")
        name = r.pop("Channel Name")
        frequency = float(r.pop("RX Freq"))
        offset = round(float(r.pop("TX Freq")) - frequency, 1)
        color_code = r.pop("Color Code")
        power = r.pop("Power")
        tg_name = r.pop("Talk Group")
        talkgroup = Talkgroup.from_contact(
            talkgroups_by_name[tg_name], r.pop("TimeSlot"),
        )
        zones.setdefault(zname, []).append(
            DigitalChannel(
                name=name,
                code=code or None,
                frequency=frequency,
                offset=offset,
                color_code=color_code,
                power=power,
                talkgroup=talkgroup,
            )
        )
    return zones


def _log_zones_channels(in_zones, log_filename=None, level=logging.DEBUG):
    filename_str = ""
    if log_filename:
        filename_str = " from {}".format(log_filename)
    total_channels = sum(len(z) for z in in_zones.values())
    logger.log(level, "Load %s zones (%s channels)%s", len(in_zones), total_channels, filename_str)


def update_zones_channels(zones_dict, in_zones, log_filename=None):
    """
    Update `zones_dict` with the contents of `in_zones`

    :param log_filename: used for logging only
    """
    _log_zones_channels(in_zones, log_filename)
    # XXX: instead, consider combining channels from same-named zones in different CSV files?
    zones_dict.update(in_zones)


def Codeplug_from_k7abd(input_dir):
    """
    :param input_dir: directory on the filesystem containing K7ABD ACB files
    :return: Codeplug
    """
    d = Path(input_dir)
    zones = {}
    talkgroups = {}
    all_talkgroups_by_name = {}
    total_files = 0
    for p in d.glob("Analog__*.csv"):
        update_zones_channels(zones, Analog_from_csv(p.read_text().splitlines()), log_filename=p)
        total_files += 1
    for p in d.glob("Talkgroups__*.csv"):
        name = p.name.replace("Talkgroups__", "").replace(".csv", "")
        talkgroups[name] = Talkgroups_map_from_csv(p.read_text().splitlines())
        logger.debug("Load %s talkgroups from %s", len(talkgroups[name]), p)
        # XXX: potential bug here if talkgroup definitions differ between files
        all_talkgroups_by_name.update(talkgroups[name])
    for p in d.glob("Digital-Others__*.csv"):
        update_zones_channels(zones, DigitalChannels_from_k7abd_csv(p.read_text().splitlines(), all_talkgroups_by_name), log_filename=p)
        total_files += 1
    for p in d.glob("Digital-Repeaters__*.csv"):
        zname = p.name.replace("Digital-Repeaters__", "").replace(".csv", "")
        # merge Talkgroup files, but prefer talkgroup names from this zone
        tg_csv = all_talkgroups_by_name.copy()
        tg_csv.update(talkgroups[zname])
        update_zones_channels(zones, {zname: tuple(DigitalRepeaters_from_k7abd_csv(p.read_text().splitlines(), tg_csv))}, log_filename=p)
        total_files += 1
    _log_zones_channels(in_zones=zones, log_filename="{} total files".format(total_files), level=logging.INFO)
    return Codeplug_from_zone_dicts(zones)



import json
import logging
import time

import attr

from dzcb.model import AnalogChannel, Contact, DigitalChannel, GroupList
import dzcb.munge

logger = logging.getLogger(__name__)

NAME_MAX = 16

value_replacements = {
    None: "None",
    False: "No",
    True: "Yes",
}

Contact_name_map = dict(
    name="Name",
    dmrid="CallID",
    kind="CallType",
)


def Contact_to_dict(c):
    d = dict(
        CallReceiveTone="No",
    )
    d.update(
        {
            Contact_name_map[k]: value_replacements.get(v, str(v))
            for k, v in attr.asdict(c).items()
            if k in attr.fields_dict(Contact)
        }
    )
    return d


GroupList_name_maps = dict(
    name="Name",
    contacts="Contact",
)


def GroupList_to_dict(g):
    return {
        "Name": g.name,
        "Contact": [tg.name for tg in g.contacts]
    }


def ScanList_to_dict(s):
    return dict(
        Name=dzcb.munge.zone_name(s.name, NAME_MAX),
        Channel=[ch.short_name for ch in s.channels],
        # Default settings
        PriorityChannel1="Selected",
        PriorityChannel2="None",
        PrioritySampleTime="2000",
        SignallingHoldTime="200",
        TxDesignatedChannel="Last Active Channel",
    )


DefaultChannel = dict(
    AdmitCriteria="Color code",
    AllowTalkaround=False,
    Autoscan=False,
    ColorCode=1,
    ContactName=None,
    CtcssDecode=None,
    CtcssEncode=None,
    DCDMSwitch=False,
    DataCallConfirmed=False,
    Decode1=False,
    Decode2=False,
    Decode3=False,
    Decode4=False,
    Decode5=False,
    Decode6=False,
    Decode7=False,
    Decode8=False,
    DisplayPTTID=False,
    EmergencyAlarmAck=False,
    EmergencySystem=None,
    GPSSystem=None,
    GroupList=None,
    InCallCriteria="Follow Admit Criteria",
    LeaderMS=False,
    LoneWorker=False,
    Privacy=None,
    PrivacyNumber=1,
    PrivateCallConfirmed=False,
    QtReverse=180,
    ReceiveGPSInfo=False,
    RepeaterSlot=1,
    ReverseBurst=False,
    RxOnly=False,
    RxRefFrequency="Medium",
    RxSignallingSystem=False,
    ScanList=None,
    SendGPSInfo=False,
    Squelch=0,
    Talkaround=False,
    Tot=120,
    TotRekeyDelay=0,
    TxRefFrequency="Medium",
    TxSignallingSystem=False,
    Vox=False,
)

AnalogChannel_name_maps = dict(
    name="Name",
    frequency="RxFrequency",
    offset="TxFrequencyOffset",
    power="Power",
    rx_only="RxOnly",
    bandwidth="Bandwidth",
    squelch="Squelch",
    tone_encode="CtcssEncode",
    tone_decode="CtcssDecode",
)


def AnalogChannel_to_dict(c):
    d = DefaultChannel.copy()
    d.update(
        {
            "ChannelMode": "Analog",
            "ScanList": dzcb.munge.zone_name(c.scanlist, NAME_MAX),
        }
    )
    d.update(
        {
            AnalogChannel_name_maps[k]: v
            for k, v in attr.asdict(c).items()
            if k in attr.fields_dict(AnalogChannel) and k in AnalogChannel_name_maps
        }
    )
    d["Name"] = c.short_name
    if d["CtcssEncode"]:
        if d["CtcssEncode"].startswith("D"):
            d["CtcssEncode"] += "N"
    else:
        d["CtcssEncode"] = "None"
    if d["CtcssDecode"]:
        if d["CtcssDecode"].startswith("D"):
            d["CtcssDecode"] += "N"
    else:
        d["CtcssDecode"] = "None"
    d["Bandwidth"] = str(round(d["Bandwidth"], 1)).replace(".0", "")
    return d


DigitalChannel_name_maps = dict(
    name="Name",
    frequency="RxFrequency",
    offset="TxFrequencyOffset",
    power="Power",
    rx_only="RxOnly",
    bandwidth="Bandwidth",
    color_code="Color Code",
)


def DigitalChannel_to_dict(c):
    d = DefaultChannel.copy()
    d.update(
        {
            "ChannelMode": "Digital",
            "RepeaterSlot": str(c.talkgroup.timeslot) if c.talkgroup else 1,
            "ContactName": str(c.talkgroup.name) if c.talkgroup else "Parrot 1",
            "GroupList": str(c.grouplist.name) if c.grouplist else None,
            "ScanList": dzcb.munge.zone_name(c.scanlist, NAME_MAX),
        }
    )
    d.update(
        {
            DigitalChannel_name_maps[k]: v
            for k, v in attr.asdict(c).items()
            if k in attr.fields_dict(DigitalChannel) and k in DigitalChannel_name_maps
        }
    )
    d["Name"] = c.short_name
    return d


Channel_value_replacements = {
    None: "None",
    False: "Off",
    True: "On",
}


def Channel_to_dict(c):
    d = None
    if isinstance(c, AnalogChannel):
        d = AnalogChannel_to_dict(c)
    elif isinstance(c, DigitalChannel):
        d = DigitalChannel_to_dict(c)
    if d is None:
        raise ValueError("Unknown type: {}".format(c))
    return {k: Channel_value_replacements.get(v, str(v)) for k, v in d.items()}


def Zone_to_dict(z):
    return {
        "Name": dzcb.munge.zone_name(z.name, NAME_MAX),
        "ChannelA": [ch.short_name for ch in z.channels_a],
        "ChannelB": [ch.short_name for ch in z.channels_b],
    }


def Codeplug_to_json(cp, based_on=None):
    cp_dict = {}
    if based_on is not None:
        cp_dict = json.load(based_on)
    # determine supported frequency range from BasicInformation
    ranges = []
    basic_info = cp_dict.get("BasicInformation", {})
    if "LowFrequency" in basic_info:
        ranges.append((basic_info["LowFrequency"], basic_info["HighFrequency"]))
    elif "LowFrequencyA" in basic_info:
        ranges.append((basic_info["LowFrequencyA"], basic_info["HighFrequencyA"]))
        ranges.append((basic_info["LowFrequencyB"], basic_info["HighFrequencyB"]))
    if ranges:
        cp = cp.filter_frequency_range(*ranges)
    cp_dict.update(
        dict(
            Contacts=[Contact_to_dict(c) for c in cp.contacts],
            Channels=[Channel_to_dict(c) for c in cp.channels],
            GroupLists=[GroupList_to_dict(c) for c in cp.grouplists],
            ScanLists=[ScanList_to_dict(c) for c in cp.scanlists],
            Zones=[Zone_to_dict(c) for c in cp.zones],
        )
    )
    # Set the programming date in intro text
    general_settings = cp_dict.setdefault("GeneralSettings", {})
    general_settings["IntroScreenLine1"] = time.strftime("%Y-%m-%d")
    logger.info(
        "Assemble JSON for %s",
        basic_info.get("Model", "Unknown. (probably won't work!)"),
    )
    return json.dumps(cp_dict, indent=2)



"""
Write series of CSV files acceptable for import into gb3gf codeplug tool

Ex: OpenGD77
"""
import csv
import logging

from dzcb.model import AnalogChannel

logger = logging.getLogger(__name__)

# These talkgroups are removed until the TG list is 32 channels or less
TALKGROUP_LIST_OVERFLOW = [
    "Michigan 1",
    "Ontario 2",
    "PS1-DNU",
    "PS2-DNU",
    "SNARS 1-2",
    "USA 2",
    "Worldwide 2",
    "TAC Eng 123",
    "WW English 2",
    "SoCal 2",
    "Audio Test 2",
]

# TG_List Overflow (These are removed if there are > 77 TG Lists)
TG_LIST_OVERFLOW = ["MMP TGS"]
TG_LIST_MAX = 76
NAME_MAX = 16

value_replacements = {
    None: "None",
    False: "No",
    True: "Yes",
}


def Codeplug_to_gb3gf_opengd77_csv(cp, output_dir):
    # filter down to supported frequency ranges
    cp = cp.filter_frequency_range((136.0, 174.0), (400.0, 480.0))
    # will keep track of contacts separately and write them at the end
    # using name_with_timeslot
    contacts = set()
    # Channels.csv, Contacts.csv, TG_List.csv, Zones.csv
    channel_fields = [
        "Channel Number",
        "Channel Name",
        "Channel Type",
        "Rx Frequency",
        "Tx Frequency",
        "Colour Code",
        "Timeslot",
        "Contact",
        "TG List",
        "RX Tone",
        "TX Tone",
        "Power",
        "Bandwidth",
        "Squelch",
        "Rx Only",
        "Zone Skip",
        "All Skip",
        "TOT",
        "VOX",
    ]
    with open("{}/Channels.csv".format(output_dir), "w", newline="") as f:
        csvw = csv.DictWriter(f, channel_fields, delimiter=";")
        csvw.writeheader()
        for ix, channel in enumerate(cp.channels):
            if isinstance(channel, AnalogChannel):
                d = {
                    "Channel Type": "Analog",
                    "RX Tone": channel.tone_decode or "None",
                    "TX Tone": channel.tone_encode or "None",
                    "Colour Code": "None",
                    "Contact": "N/A",
                    "TG List": "None",
                }
            else:
                d = {
                    "Channel Type": "Digital",
                    "RX Tone": "None",
                    "TX Tone": "None",
                    "Colour Code": channel.color_code,
                    "Contact": "N/A",
                    "TG List": channel.grouplist.name if channel.grouplist else "None",
                }
                if channel.talkgroup:
                    d["Contact"] = channel.talkgroup.name_with_timeslot
                    contacts.add(channel.talkgroup)
            d.update(
                {
                    "Channel Number": ix + 1,
                    "Channel Name": channel.short_name,
                    "Rx Frequency": channel.frequency,
                    "Tx Frequency": round(channel.frequency + channel.offset, 5),
                    "Timeslot": 1,
                    "Power": str(channel.power),
                    "Bandwidth": "25KHz" if channel.bandwidth > 19 else "12.5KHz",
                    "Squelch": str(channel.squelch) if channel.squelch else "Disabled",
                    "Rx Only": value_replacements[channel.rx_only],
                    "Zone Skip": "No",
                    "All Skip": "No",
                    "TOT": 90,
                    "VOX": "No",
                }
            )
            csvw.writerow(d)
    tg_fields = ["TG List Name"] + ["Contact {}".format(x) for x in range(1, 33)]
    with open("{}/TG_Lists.csv".format(output_dir), "w", newline="") as f:
        csvw = csv.DictWriter(f, tg_fields, delimiter=";")
        csvw.writeheader()
        n_grouplists = len(cp.grouplists)
        for gl in cp.grouplists:
            if n_grouplists > TG_LIST_MAX and gl.name in TG_LIST_OVERFLOW:
                n_grouplists -= 1
                continue
            tg_list = {"TG List Name": gl.name}
            contacts_by_name = {tg.name: tg for tg in gl.contacts}
            remove_tgs = list(reversed(TALKGROUP_LIST_OVERFLOW))
            # remove some talkgroups to get under the limit
            while len(contacts_by_name) > 32:
                try:
                    del contacts_by_name[remove_tgs.pop()]
                except KeyError:
                    pass
            for ix, tg in enumerate(contacts_by_name.values()):
                tg_list["Contact {}".format(ix + 1)] = tg.name_with_timeslot
                contacts.add(tg)
            csvw.writerow(tg_list)
    zone_fields = ["Zone Name"] + ["Channel {}".format(x) for x in range(1, 81)]
    with open("{}/Zones.csv".format(output_dir), "w", newline="") as f:
        csvw = csv.DictWriter(f, zone_fields, delimiter=";")
        csvw.writeheader()
        zone_names = [z.name for z in cp.zones]
        # OpenGD77 doesn't have scanlist, so simulate it with separate zones
        write_zones = [sl for sl in cp.scanlists if sl.name not in zone_names]
        write_zones.extend(cp.zones)
        for zone in write_zones:
            row = {"Zone Name": zone.name}
            for ix, ch in enumerate(zone.unique_channels):
                if ix + 1 > 80:
                    logger.debug("Zone '%s' exceeds 80 channels", zone.name)
                    break
                row["Channel {}".format(ix + 1)] = ch.short_name
            csvw.writerow(row)
    with open("{}/Contacts.csv".format(output_dir), "w", newline="") as f:
        csvw = csv.DictWriter(
            f, ["Contact Name", "ID", "ID Type", "TS Override"], delimiter=";"
        )
        csvw.writeheader()
        for tg in sorted(contacts, key=lambda c: c.name_with_timeslot):
            csvw.writerow(
                {
                    "Contact Name": tg.name_with_timeslot,
                    "ID": tg.dmrid,
                    "ID Type": str(tg.kind),
                    "TS Override": str(tg.timeslot),
                }
            )
    logger.info("Wrote GB3GF OpenGD77 CSV files to '%s'", output_dir)



import codecs
from glob import glob
from os.path import abspath, basename, dirname, join, splitext

import setuptools


def read(*parts):
    """Read a file in this repository."""
    here = abspath(dirname(__file__))
    with codecs.open(join(here, *parts), 'r') as file_:
        return file_.read()


setuptools.setup(
    name='dzcb',
    use_scm_version=True,
    description='DMR Zone Channel Builder',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author='Masen Furer',
    author_email='m_github@0x26.net',
    url='https://github.com/masenf/dzcb',
    package_dir={"": 'src'},
    package_data={'dzcb': ['data/*.json', 'data/*.csv', 'data/k7abd/*.csv', 'data/farnsworth/*.json']},
    packages=setuptools.find_packages('src'),
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    python_requires='>=3.5,<4',
    setup_requires=[
        'setuptools_scm >= 3.3',
    ],
    install_requires=[
        'appdirs==1.4.4',
        'attrs',
        'geopy==2.1.0',
        'importlib-resources',
        'requests~=2.24.0',
    ],
    classifiers=[
        'Intended Audience :: Developers',
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
)



"""
dzcb.contacts_trim - remove contacts to get under radio limits
"""
import sys

remove_suffixes = [",", ",GR", ",CY", ",CN", ",BE", ",FR", ",ES", ",IT", ",RU", ",PL", ",DE", ",PT", ",TR", ",SI", ",JP", ",Korea Republic of", ",PH", ",MY", ",TH", ",AR", ",BR", ",CL", ",CO", ",VE", ",UY", ",SE", ",CH", ",CZ", ",SK", "BA", "HR", "AT"]

total = 0

def check_suffix(line):
    for s in remove_suffixes:
        if line.strip().endswith(s):
            return None
    return line

for line in sys.stdin:
    outline = check_suffix(line)
    if outline:
        total += 1
        sys.stdout.write(outline)

sys.stderr.write(f"Wrote {total} records")



import logging
from pathlib import Path
import time

logger = logging.getLogger(__name__)
LOG_FORMAT = "%(asctime)s [%(name)s] [%(levelname)s]  %(message)s"


def change_log_level(to):
    def _change_log_level(r):
        r.levelno = to
        r.levelname = logging.getLevelName(r.levelno)
        return True

    return _change_log_level


def ltrim_to_delimiter(d):
    def _ltrim_to_delimiter(r):
        r.args = (r.args[0].partition(d)[2], *r.args[1:])
        return True

    return _ltrim_to_delimiter


def keep_to_delimiter(d):
    def _keep_to_delimiter(r):
        r.args = (r.args[0].partition(d)[0], *r.args[1:])
        return True

    return _keep_to_delimiter


def init_logging(log_path=None):
    formatter = logging.Formatter(LOG_FORMAT)
    root = logging.getLogger()
    root.setLevel(logging.NOTSET)

    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(logging.INFO)
    root.addHandler(stream_handler)

    if log_path is not None:
        filename = time.strftime("dzcb.%Y_%m_%d_%H%M%S.log")
        path = Path(log_path) / filename
        logger.info("Logging to '%s' at DEBUG level", path)
        file_handler = logging.FileHandler(path)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.DEBUG)
        root.addHandler(file_handler)

        # if a file was specified, spew warnings there instead of the console
        warning_logger = logging.getLogger("py.warnings")
        warning_logger.addFilter(filter=change_log_level(to=logging.DEBUG))
        warning_logger.addFilter(filter=ltrim_to_delimiter(": "))
        warning_logger.addFilter(filter=keep_to_delimiter("\n"))
        logging.captureWarnings(True)
